#############################################################################
# ClearCase/ClearPrompt.pm -- ClearCase/Perl interface to clearprompt(1)
#
# Copyright (C) 1999-2000 David Boyce. All rights reserved.
#
# This file is part of "ClearPerl" version 5 (ClearPerl5).
# This file/module is free software; you can redistribute it and/or
# modify it under the same terms as ClearPerl itself, as stated in
# the README file of the ClearPerl distribution.
#############################################################################

package ClearCase::ClearPrompt;

require 5.004;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
require Exporter;
$VERSION = 5.000_013;
@ISA = qw(Exporter);
@EXPORT = qw(clearprompt);

use ClearCase::Config qw(clearcase_progpath);
use ClearCase::OSUtil qw(MSWIN &tmpname);

my $CPT = undef;

# Run clearprompt with specified args and return what it returns. Uses the
# exact same syntax as the clearprompt executable ('ct man clearprompt')
# except the -outfile flag, which is handled internally here.
sub clearprompt(@)
{
   my $mode = shift;
   my @args = @_;

   local $!;	# don't mess up errno in the caller's world.

   # Aid in finding clearprompt.
   $CPT ||= clearcase_progpath('clearprompt');

   # On Windows we must add an extra level of escaping to the prompt string
   # since all forms of system() appear to go through the cmd.exe shell!
   if (MSWIN) {
      for my $i (0..$#args) {
	 if ($args[$i] =~ /^-pro/) {
	    $args[$i+1] =~ s/"/'/gs;
	    $args[$i+1] = qq("$args[$i+1]");
	    last;
	 }
      }
   }

   # For clearprompt modes in which we get textual data back via a file,
   # derive here a reasonable temp-file name and handle the details
   # of reading data back out of it and unlinking it when done.
   # For other modes, just fire off the cmd and return the status.
   # In a void context, don't wait for the button to be pushed; just
   # "fork" and proceed asynchonously since this is presumably just an
   # informational message.
   # If the cmd took a signal, return undef and leave the signal # in $?.
   if ($mode =~ /text|file|list/) {
      my $outf = tmpname($mode);
      my $data;
      if (!system($CPT, $mode, '-out', $outf, @args)) {
	 if (open(OUTFILE, $outf)) {
	    local $/ = undef;
	    $data = <OUTFILE>;
	    $data = '' if !defined $data;
	    close(OUTFILE);
	 }
      } else {
	 # If we took a signal, return undef with the signo in $?. The
	 # clearprompt cmd apparently catches SIGINT and returns 0x400 for
	 # some crazy reason; we fix it here so $? looks like a normal sig2.
	 $? = 2 if $? == 0x400;  # see above
	 $data = undef if $? && $? <= 0x80;
      }
      unlink $outf if -f $outf;
      return $data;
   } else {
      # Never proceed asynchronously in a command-line env ...
      if (defined wantarray || !$ENV{ATRIA_FORCE_GUI}) {
	 system($CPT, $mode, @args);
	 $? = 2 if $? == 0x400;  # see above
	 return ($? && $? <= 0x80) ? undef : $?>>8;
      } else {
	 if (MSWIN) {
	    system(1, $CPT, $mode, @args);
	 } else {
	    return if fork;
	    exec($CPT, $mode, @args);
	 }
      }
   }
}

1;
__END__

=head1 NAME

ClearCase::ClearPrompt - Handle clearprompt in a portable, convenient way

=head1 SYNOPSIS

    use ClearCase::ClearPrompt qw(clearprompt);

    # boolean usage
    my $rc = clearprompt(qw(yes_no -mask y,n -type ok -prompt), 'Well?');

    # returns text into specified variable (context sensitive).
    my $txt = clearprompt(qw(text -pref -prompt), 'Enter text data here');

    # asynchronous usage - show dialog box and continue
    clearprompt(qw(proceed -mask p -type ok -prompt), "You said: $txt");

=head1 DESCRIPTION


Native ClearCase provides a utility (B<clearprompt>) for collecting
user input or displaying messages within triggers. However, usage of
this tool is awkward and error-prone, especially in multi-platform
environments.  Often you must create temp files, invoke clearprompt to
write into them, open them and read the data, then unlink them. In many
cases this code must run seamlessly on both Unix and Windows systems
and is replicated throughout many trigger scripts. ClearCase::ClearPrompt
abstracts this dirty work without changing the interface to
B<clearprompt> at all.

The C<clearprompt()> function takes the exact same set of flags as
the eponymous ClearCase command, e.g.:

    my $response = clearprompt('text', '-def', '0', '-pro', 'Well? ');

except that the C<-outfile> flag is unnecessary since creation,
reading, and removal of this temp file is managed internally.

In a void context, clearprompt() behaves asynchronously; i.e. it
displays the dialog box and returns so that execution can continue.  In
any other context it waits for the dialog's button to be pushed and
returns the appropriate data type.

The clearprompt() I<function> always leaves the return code of the
clearprompt I<command> in C<$?> just as C<system()> would.  If the
prompt was interrupted via a signal, the function returns the undefined
value.

-head1 NOTE

I<An apparent undocumented "feature" of clearprompt(1) is that it
catches SIGINT (Ctrl-C) and provides a status of 4 rather than
returning the signal number in C<$?> according to normal (UNIX) signal
semantics.>  We fix that up here so it looks like a normal signal 2.
Thus, if C<clearprompt()> returns undef the signal number is reliably
in $? as it should be.


=head1 PORTING

This package is known to work fine on Solaris 2.5.1 and Windows NT
4.0.  As these two platforms are quite different this should take care
of any I<significant> portability issues, but please send reports of
tweaks needed for other platforms to the address below.

=head1 AUTHOR

David Boyce <dsb@world.std.com>

Copyright (c) 1999-2000 David Boyce. All rights reserved.  This perl program
is free software; you may redistribute it and/or modify it under the
same terms as Perl itself.

=head1 SEE ALSO

clearprompt(1), ClearCase::OSUtil, perl(1)

=cut

1;

__END__
