#############################################################################
# ClearCase/ClearTool.pm -- base class/package for 'cleartool' commands
#
# Copyright (C) 1995-2000 Brad Appleton. All rights reserved.
#
# This file is part of "ClearPerl" version 5 (ClearPerl5).
# This file/module is free software; you can redistribute it and/or
# modify it under the same terms as ClearPerl itself, as stated in
# the README file of the ClearPerl distribution.
#############################################################################

package ClearCase::ClearTool;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
use vars qw($AUTOLOAD);
use vars qw($DEFAULT_CLEARTOOL %opts @argv);

use Exporter;
use Carp;

use ClearCase::Config qw(clearcase_config);
use ClearCase::ExecOpts qw(execopts_parse execopts_cmdline);

@ISA = qw(Exporter AutoLoader);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

@EXPORT = qw( &cleartool );
@EXPORT_OK = qw( 
                 &cleartool_output
                 &cleartool_status
                 &cleartool_success
                 &cleartool_open
               );
%EXPORT_TAGS = ( 'all' => [@EXPORT, @EXPORT_OK] );

$VERSION = 5.000_013;
require  5.004;    ## requires this Perl version or later

#############################################################################

=head1 NAME

ClearCase::ClearTool, cleartool - Interface for executing cleartool(1)
subcommands.

=head1 SYNOPSIS

  use ClearCase::ClearTool;

  ## Use an *explicit* 'ClearTool' object to execute 'cleartool' commands
  my $ct = new ClearCase::ClearTool;
     ## get help
  $ct->help;
     ## different syntax for the same 'lsco' invocation
  $ct->lsco(-recurse, -me, -short, '/vobs/myproj/src');
  $ct->lsco( qw(-recurse -me -short /vobs/myproj/src) );

  ## Use an *implicit* 'ClearTool' object to execute 'cleartool' commands
     ## get help
  cleartool 'help';
     ## different syntax for the same 'lsco' invocation
  cleartool qw(lsco -recurse -me -short /vobs/myproj/src);
  cleartool 'lsco', -recurse, -me, -short, '/vobs/myproj/src';
  cleartool 'lsco', qw(-recurse -me -short /vobs/myproj/src);
  cleartool->lsco(-recurse, -me, -short, '/vobs/myproj/src');
  cleartool->lsco( qw{-recurse -me -short /vobs/myproj/src} );

=head1 REQUIRES/IMPORTS

perl5.004, ClearCase::Config

=head1 EXPORTS

By default, B<cleartool> is exported. The functions B<cleartool>,
B<cleartool_output>, B<cleartool_status>, B<cleartool_success>
and B<cleartool_open> may be exported upon explicit request.

=head1 DESCRIPTION

B<ClearCase::ClearTool> is both a module and a class. As a class, it defines
an abstract baseclass and interface for objects that correspond to the
C<cleartool> command that is part of ClearCase. A C<ClearTool> object
may be created for executing C<cleartool> subcommands.

As a module, B<ClearCase::ClearTool> provides a function B<cleartool()>
that may also be used to execute C<cleartool> subcommands. The difference
between executing subcommands using the B<cleartool()> function versus
invoking methods on a C<ClearTool> object is that the former uses an
implicit (or default) C<ClearTool> object while the latter uses an
object explicitly created by the user.

=head1 EXECUTION OPTIONS AND RETURN VALUES

By default, the result of invoking a C<cleartool> subcommand method
(either implicitly or explicitly) of a C<ClearTool> object will return
the standard output of the subcommand:

  my $output = cleartool->ls;

The returned result is a single string corresponding to what the
subcommand printed on its standard output (by default, the trailing
newline is removed).

If invoked in an array context, rather than returning a lone string,
an array of strings is returned (one per line of output):

  my @lines = cleartool->ls;

Each element of the array C<@output> will, by default, have any trailing
newline removed before the result is returned to the user.

For the B<cleartool()> function, there is a universal exception
to this rule: Whenever B<cleartool()> is invoked with I<no arguments>,
it I<ALWAYS> returns a reference to the implicit C<ClearTool> object
it is currently using! This makes sense because if no arguments are
given, then no (sub)command was executed (so there is no output or
exit status to examine). In fact, this is precisely what allows us to
invoke B<cleartool()> using both procedural I<and> object syntax, as in:

  my $output = cleartool 'ls';
  $output = cleartool->ls;

The second invocation given above actually invokes the B<ls()> method
of the object returned by invoking B<cleartool()> with zero arguments!

Now, back to our story ....

The removal of trailing newlines may be disabled and (re)enabled by
specifying the B<-nochomping> option when invoking the B<cleartool()>
function, or when invoking a subcommand:

  ## Disable "chomping" of newlines, only lasts for a single invocation
  my $output = cleartool ['-nochomping'], 'ls';

  ## Same as above using an explicit ClearTool object
  $output = cleartool->ls( ['-nochomping'] );

  ## Same as above using an explicit ClearTool object
  $output = $ct->ls( ['-nochomping'] );

The B<-nochomping> keyword is an example of an I<exec-option>: an option
that controls the execution or return value of B<cleartool>. Other
exec-options can be used to make the return value be the exit-status
instead of the subcommand output; a filehandle from which to read the
output of the subcommand; or even a boolean value corresponding to whether
or not the command returned a 0 exit status for "success." See
L<ClearCase::ExecOpts> for a complete description of the name
and meaning of all possible exec-options.

=head1 SPECIFYING EXEC-OPTIONS FOR SUBCOMMANDS

For a given B<cleartool> subcommand, exec-options which override any
defaults for the subcommand invocation may be specified by giving a list
of arguments similar to the list accepted by an B<ExecOpts> constructor,
or else an explicit B<ExecOpts> object may be specified.

B<Note that exec-options must always precede any subcommand options!> The
exec-options may be passed as an array-ref, in which case the array should
look much the same as the argument list to the B<ClearCase::ExecOpts>
constructor. It may also be passed as a hash-ref I<IF> it was the
return result from the B<execopts_parse> function! Or it can be passed
as a reference to an explicit B<ClearCase::ExecOpts> object. Finally,
as a special case, if only one exec-option is needed and the option
itself takes no arguments, then it may optionally be specified as the
first parameter to a subcommand I<provided that> it uses a '+' suffix
instead of '-'. Hence, the following are all equivalent:

   cleartool ['-status'], qw(lsco -short);
   cleartool ClearCase::ExecOpts->new('-status'), qw(lsco -short);
   cleartool qw(+status lsco -short);

   cleartool->lsco(['-status'], -short);
   cleartool->lsco(ClearCase::ExecOpts->new('-status'), '-short');
   cleartool->lsco('+status -short');

Any exec-options that are not explicitly specified in a B<cleartool>
subcommand invocation will default to the settings in the B<ExecOpts>
object that is associated with the corresponding B<ClearTool> object.

=begin _NOT_IMPLEMENTED_

Each B<ClearTool> object has a an B<ExecOpts> object associated with it
which may be obtained using the B<exec_opts()> method.

To change the actual B<ExecOpts> object associated with a B<ClearTool>
object, use the B<exec_opts()> method and supply the new B<ExecOpts>
object.

So you are free to use exec-options to mix and match the return value type
to use the syntax which seems most appropriate at any given time. It is
very easy to go overboard with this feature; I<Don't>! It is there as a
convenience, but for the sake of readability, its use should probably
be kept to a minimum, instead preferring a more consistent syntax and
explicitly altering it (temporarily) only when it enhances program
readability and comprehension.

Just remember that explicit exec-options given to a B<cleartool>
subcommand take effect I<only for the immediate invocation>; but modifying
or resetting exec-options via the B<exec_opts()> method will continue to
be in effect for all subsequently executed subcommands (unless explicitly
overridden for a single invocation, or until explicitly reset).

=end _NOT_IMPLEMENTED_

=cut

#############################################################################

=head1 CONSTRUCTION/CREATION METHODS

The following methods are used to create or initialize a B<ClearTool> object.

=cut

##---------------------------------------------------------------------------

=head1 new()

Create a new B<ClearTool> object.

    ## Create a new ClearTool object for executing subcommands
    my $ct1 = ClearCase::ClearTool->new;

    ## Same as above, but used a different syntax
    my $ct2 = new ClearCase::ClearTool;

    ## Create new ClearTool object, and have the default return
    ## value for subcommands be a reference to the object rather
    ## than the output.
    my $ct3 = ClearCase::ClearTool->new( '+NoChomping' );

    ## Create new ClearTool object which uses the 'cleartool' executable
    ## residing in the specified directory.
    my $ct4 = new ClearCase::ClearTool( '+ExecPath' => '/site_local/bin' );

=cut

sub new {
    ## Determine if we were called via an object-ref or a classname
    my $this = shift;
    my $class = ref($this) || $this;
    ## Any remaining arguments are treated as initial values for the
    ## hash that is used to represent this object.
    my $ctpath = clearcase_config('CLEARTOOL');
    my @defaults = ( qw( -output -chomping ),
                     '-abort'    => undef,
                     '-execpath' => $ctpath
                   );
    my $execopts = execopts_parse(@defaults, @_);
    my $self = { '-execopts' => $execopts };
    ## Bless ourselves into the desired class and perform any initialization
    bless $self, $class;
    $self->initialize();
    return $self;
}

##---------------------------------------------------------------------------

sub initialize {
    my $self = shift;
}

#############################################################################
##                           UTILITY ROUTINES
#############################################################################

## _getopts is used to "parse" all exec-options from the argument list.
## It returns a hashref corresponding to the exec-options, and modifies
## the given array to shift pass the exec-options to the first argument.

sub _getopts( $ \@ ) {
    my $self = shift;
    local *argv = shift;

    ## If the first element of argv is a hashref, we must have already
    ## computed the exec-opts and the 1st arg was the result
    return  shift(@argv)  if (ref($argv[0]) eq 'HASH');

    ## Otherwise, all initial arguments that are refs, or else begin with
    ## a '+' are exec-opts to be parsed
    my @execopts = ();
    while (ref $argv[0]  or  substr($argv[0], 0, 1) eq '+') {
       push @execopts, shift(@argv);
    }

    ## Now get defaults for current object and parse the given options
    ## and return the combined results
    my $defaults = $self->{'-execopts'};
    execopts_parse($defaults, \@execopts);
}


## _default_cleartool returns the default ClearTool object (creating it
## if it doesnt already exist). Its basically the "accessor" to a singleton
## objbect that serves as the default when no other object was given.

$DEFAULT_CLEARTOOL = undef;

sub _default_cleartool {
   unless (defined $DEFAULT_CLEARTOOL) {
      $DEFAULT_CLEARTOOL = ClearCase::ClearTool->new();
   }
   return  $DEFAULT_CLEARTOOL;
}

#############################################################################

=head1 OBJECT METHODS

The following methods may be invoked for a constructed B<ClearTool> object.

=cut

##---------------------------------------------------------------------------

=head1 output()

Return standard output of the specified subcommand

    my $output_str = $ct->output('subcmd', $arg1, $arg2, @rest);
    my @output_lines = $ct->output("subcmd $arg1 $arg2 @rest");

In a scalar context, this method executes the specified subcommand
and returns the output as a single string. In an array context,
it returns an array of output lines. If you wish to control whether or not
the return value has trailing newlines removed, look at the B<-chomping>
and B<-nochomping> exec-options.

=cut

sub output { 
   my $self = shift;

   ## Look at given options + current defaults and determine
   ## the command-line invocation to use
   local *opts = _getopts($self, @_);
   my $cmdstr  = execopts_cmdline(\%opts, @_);

   ## Execute the command and capture the output
   my $wantarray = wantarray();
   my ($outstr, @outarray) = ( "", () );
   if ($wantarray) {
      @outarray = qx($cmdstr);
   }
   else {
      $outstr = qx($cmdstr);
   }

   ## Save status, and die on failure if -abort was set
   my $status = $self->{'-status'} = $?;
   if ($status  and  $opts{'-abort'}) {
       $self->abort('-abort' => $opts{'-abort'}, '-cmdstr' => $cmdstr);
   }

   ## Otherwise chomp as necessary
   $opts{'-chomping'}  and  chomp($wantarray  ?  @outarray  :  $outstr);

   ## Return the result
   return  $wantarray  ?  @outarray  :  $outstr;
}

##---------------------------------------------------------------------------

=head1 status()

Execute a subcommand and return its status, or else the exit-status of
the last subcommand executed.

    ## Execute a subcommand and return its status
    $exit_status = $ct->status("subcmd $arg1 $arg2 @rest");

    ## Same thing, but specify the subcommand as a list
    $exit_status = $ct->status('subcmd', $arg1, $arg2, @rest);

    ## Get status of last subcommand executed
    my $exit_status = $ct->status();

The value returned will be similar to that returned by the builtin
B<system> function, and contained in the C<$?> variable.

I<NOTE> that this method I<ignores> the B<-abort> exec-option! If auto-abort
handling was performed, then the exit-status would only ever be one value 
upon a successful subcommand execution. Therefore it is assumed that since
the exit-status is explicitly requested, the caller would prefer to perform
their own "handling" based upon the status returned.

=cut

sub status {
   my $self = shift;
   if (@_ == 1  and  $_[0] =~ /^[-+]?\d+$/) {
       ## A lone status-value was given to set the status;
       ## set stored-status to the given value and return
       $self->{-status} = shift;
   }
   ## If no remaining arguments, just return current status
   return  $self->{-status}  if (@_ == 0);

   ## We have some arguments so we need to run a subcommand first.
   ## Look at given options + current defaults and determine the
   ## command-line invocation to use
   local *opts = _getopts($self, @_);
   my $cmdstr = execopts_cmdline(\%opts, @_);

   ## Execute the command and save the status;
   my $status = $self->{-status} = system( $cmdstr );

   ## Return the status (note that we dont bother with '-abort')
   return  $status;
}

##---------------------------------------------------------------------------

=head1 success()

Return true if the last subcommand returned an exit-value of zero. 

    if ($ct->success) {
       print "Last executed subcommand succeeded!\n";
    }
    else {
       print "Last executed subcommand failed!\n";
    }

This method will compare the exit-status of the last executed subcommand
against a value of zero (C<0>). If the comparison is equal, then a
true (non-zero) value will be returned. Otherwise a false value (0 or "")
will be returned.

=cut

sub success {
    my $self = shift;
    return  ($self->status(@_) == 0);
}

##---------------------------------------------------------------------------

=head1 open()

Return an ClearTool::Handle object for reading the output of a subcommand

   my $fhandle = $ct->open("subcmd $arg1 $arg2 @rest");
   while (<$fhandle>) { ... }
   my @lines = <$fhandle>;
   $fhandle->close();

This method returns a file handle that may be used to read the output of
the given subcommand one line at a time. B<close()> should be called when
done reading to obtain the exit status of the command (using
B<status()> or B<success()> on the ClearTool object), or to trigger
B<-abort> actions.

=cut

sub open {
   my $self = shift;

   ## Look at given options + current defaults and determine
   ## the command-line invocation to use
   local *opts = _getopts($self, @_);

   my $cmdstr  = execopts_cmdline(\%opts, @_);

   ## get new ClearTool filehandle and return it
   ClearCase::ClearTool::Handle->new( $self, $cmdstr, $opts{'-abort'} );
}

##---------------------------------------------------------------------------

## 'execute' is like 'cleartool()' but it is a method rather than a
## function. Actually, thats not quite true. It *does* expect a ClearTool
## object as its first argument, HOWEVER, for performance reasons, it is
## NOT invoked using method syntax by the AUTOLOAD routine; it is invoked
## as a plain old function using 'goto &execute'.

sub execute (@) {
   my $self = shift;
   return  $self  if (@_ == 0);

   ## Get options
   local *opts = _getopts($self, @_);

   ## Next argument is the subcommand and anything afterward is arguments
   ## for the subcommand. So we need to dispatch to the proper function
   ## cased on the desired return type. The name of the return-type
   ## should correspond to the name of the method to invoke (one of:
   ## 'output', 'status', 'success', 'open').
   my $return_type = $opts{'-returns'};

   ## guard against crash: check we're calling the right thing
   die( "execute usage error (-returns => $return_type?)\n" )
     unless $return_type eq 'output'  ||
            $return_type eq 'status'  ||
            $return_type eq 'success' ||
            $return_type eq 'open';
   $self->$return_type(\%opts, @_);
}

## 'cmdline ' is like 'execopts_cmdline' except it is a method instead
## of a function
sub cmdline {
   my $self = shift;

   ## Look at given options + current defaults and determine
   ## the command-line invocation to use
   local *opts = _getopts($self, @_);
   return  execopts_cmdline(\%opts, @_);
}

## abort is used to perform autoabort handling. If the '-abort'
## exec-option is turned on, then _abort is called when $ct->output()
## or $ct->open() fails. It needs to know the command-line that
## misexecuted, whether it was for output() or open(), the return
## status and any (possibly default) error message. It also needs to
## know if a user-supplied "die" and/or message-string should be used
## (and it should be smart enough to peek at $? and $!).

sub abort( @ ) {
   my $self = shift;
   ### Get abort "flags"
   my %flags  = @_;
   my $cmdstr = $flags{'-cmdstr'} || "";
   my $abort  = $flags{'-abort'}  || undef;

   ## Set default error subroutine & message
   my $abort_sub = sub { die @_; };
   my $abort_msg = "command failed - exit-status is $?";
   $abort_msg .= " ($!)"  if ($!);
   $abort_msg .= ":\n\t$cmdstr"  if ($cmdstr);

   my $reftype   = ref $abort;
   if ($reftype eq 'HASH') {
      exists $abort->{'-errsub'} and $abort->{'-errsub'}
                                 and $abort_sub = $abort->{'-errsub'};
      exists $abort->{'-errmsg'} and $abort_msg = $abort->{'-errmsg'};
   }
   elsif ($reftype eq 'CODE') {
      $abort_sub = $abort;
   }
   elsif ((! $reftype) and (defined $abort)) {
      ## if $abort is a number, use defaults; otherwise it's the message text
      $abort_msg = $abort  unless ($abort =~ /^\d+$/);
   }

   ## Invoke the error sub with the given message
   chomp($abort_msg);
   &{ $abort_sub }( "$abort_msg\n" );
}

##---------------------------------------------------------------------------

=begin _NOT_IMPLEMENTED_

=head1 exec_opts()

Return (or reset) the B<ExecOpts> object corresponding to this B<ClearTool>
object.

    my $opts = $ct->exec_opts;
    my $is_chomping = $opt->getopts('-chomping');
    if ($is_chomping) {
        $opt->setopts('-nochomping');
    }

If no arguments are given, then a reference to the corresponding B<ExecOpts>
object is returned. If a single argument is given, it is assumed to be
a new B<ExecOpts> object to associate with this B<ClearTool> object.

=end _NOT_IMPLEMENTED_

=cut

sub exec_opts {
   my $self = shift;
   croak "NOT YET IMPLEMENTED!\n";
}

##---------------------------------------------------------------------------

#############################################################################

=head1 IMPORTABLE FUNCTIONS

By default, B<ClearCase::ClearTool> exports nothing, however several
functions with a "cleartool" prefix may be imported if desired. Each
one of the importable functions behaves much like a corresponding
object method, but uses and implicit default B<ClearTool> object.

=cut

##---------------------------------------------------------------------------

=head1 cleartool()

Execute a B<cleartool> subcommand or return the "default" B<ClearTool> object.

   my $output = cleartool 'ls', '-long';
       ## same as above, but different syntax
   $output = cleartool qw(ls -long);

       ## obtain the corresponding object and run a subcommand
   my $default_ct = cleartool();
   $output = $default_ct->pwv('-short');
       ## same as above, but in a single statement
   $output = cleartool->pwv('-short');

The B<cleartool()> function will treat its parameters as a subcommand
name and its arguments (including any exec-options and subcommand options).
It will then execute()the subcommand and return the result, as qualified
by the specified exec-options for the default B<ClearTool> object.
If no arguments are given, then B<cleartool()> returns a reference to
the default B<ClearTool> object.

=cut

sub cleartool (@) {
   my $self = _default_cleartool;
   return  (@_ == 0)  ?  $self  :  $self->execute(@_);
}

##---------------------------------------------------------------------------

=head1 cleartool_output()

Like B<cleartool()>, but I<always> returns output via the B<output()> method.

=cut

sub cleartool_output (@) {
   my $self = _default_cleartool;
   return  (@_ == 0)  ?  $self  :  $self->output(@_);
}

##---------------------------------------------------------------------------

=head1 cleartool_status()

Like B<cleartool()>, but I<always> returns status via the B<status()> method.

=cut

sub cleartool_status (@) {
   my $self = _default_cleartool;
   return  (@_ == 0)  ?  $self  :  $self->status(@_);
}

##---------------------------------------------------------------------------

=head1 cleartool_success()

Like B<cleartool()>, but I<always> returns success via the B<success()> method.

=cut

sub cleartool_success (@) {
   my $self = _default_cleartool;
   return  (@_ == 0)  ?  $self  :  $self->success(@_);
}

##---------------------------------------------------------------------------

=head1 cleartool_open()

Like B<cleartool()>, but I<always> returns a handle via the B<open()> method.

=cut

sub cleartool_open (@) {
   my $self = _default_cleartool;
   return  (@_ == 0)  ?  $self  :  $self->open(@_);
}

#############################################################################

## AUTOLOAD is used primarily for allowing subcommand-names to
## be treated as methods of a B<ClearTool> object. They always get
## translated into calls to B<execute()>.

sub AUTOLOAD {
   my $self = shift;
   (my $subcmd = $AUTOLOAD) =~ s/.*:://;
   return if $subcmd eq 'DESTROY';
   no strict 'refs';
   # install a new method '$subcmd' to avoid autoload next time ...
   *$subcmd = sub {
      ## Insert the subcmd immediately *after* any exec-options but
      ## before all other arguments
      my $lastopt = 0;
      while ( defined $_[$lastopt]
	      and ( ref $_[$lastopt]
		    or $_[$lastopt] =~ m/^\+/ ) ) {
         ++$lastopt;
      }
      splice( @_, $lastopt, 0, $subcmd );
      ## Execute the subcommand!
      goto &execute;
   };
   # ... then service this request
   return $self->$subcmd( @_ );
};

sub DESTROY { }

#############################################################################

package ClearCase::ClearTool::Handle;

use vars qw(@ISA @EXPORT @EXPORT_OK);
use vars qw( %FH );

use Exporter;

use IO::Handle;

@ISA = qw( IO::Handle );

@EXPORT = qw();
@EXPORT_OK = qw();

# Extends IO::Handle by overwriting new() and close() to achieve the
# -abort semantics.

# constructor new( cleartool-object, command-line, abort-option ):
# - Call new() of parent class
# - Save some values we'll need for the close/abort.
#
sub new{
   my( $class, $cleartool, $cmdstr, $abort ) = @_;
   my $fh = $class->SUPER::new();

   # re-bless
   bless( $fh, $class );

   # save stuff needed for close/abort
   local *FH = *$fh;
   $FH{'-cleartool'} = $cleartool;
   $FH{'-abort'}     = $abort;
   $FH{'-cmdstr'}    = $cmdstr;

   open( $fh, "$cmdstr |" )
     || die( "cannot open pipe to read from: $cmdstr ($! $?)" );
   return $fh;
}


# method close:
# - Close the handle
# - Conditionally do abort actions
#
sub close($){
   my $fh = shift();
   local *FH = *$fh;
   my $abort     = $FH{'-abort'};
   my $cmdstr    = $FH{'-cmdstr'};
   my $cleartool = $FH{'-cleartool'};

   my $res = $fh->SUPER::close();   

   ## save the status
   my $status = $cleartool->{'-status'} = $?;

   ## Die on failure if -abort was set
   if( $status  and  $abort ) {
       $cleartool->abort( '-abort'  => $abort,
                          '-cmdstr' => $cmdstr );
   }
   return $res;
}

#############################################################################

=head1 AUTHOR

Brad Appleton  F<E<lt>bradapp@enteract.comE<gt>>

=head1 SEE ALSO

ClearCase::ExecOpts,
ClearCase::Config

=cut

package ClearCase::ClearTool;

1;

__END__
