#############################################################################
# ClearCase/Config.pm -- ClearCase Configuration Settings
#
# Copyright (C) 1995-2000 Brad Appleton. All rights reserved.
#
# This file is part of "ClearPerl" version 5 (ClearPerl5).
# This file/module is free software; you can redistribute it and/or
# modify it under the same terms as ClearPerl itself, as stated in
# the README file of the ClearPerl distribution.
#############################################################################


=head1 NAME

ClearCase::Config - ClearCase configuration settings

=head1 SYNOPSIS

  use ClearCase::Config;
  my $ct_path = clearcase_config 'CLEARTOOL';
  my ($cc_home, $cc_adm) = clearcase_config('ATRIA_HOME', 'ATRIA_ADM');
  my $clearbug_path = clearcase_progpath 'clearbug';
  my $ct_version = clearcase_version 'cleartool';


=head1 REQUIRES/IMPORTS

perl5.004

=head1 EXPORTS

By default, I<nothing> is exported. The functions
B<clearcase_config>, B<clearcase_progpath>,
B<clearcase_version>, and B<clearcase_patches>
may all be exported upon request.
Functions with the almost the same names as these four, but with a leading
prefix of "_" (a single underscore) may be also be used in conjunction
with C<ClearCase::SiteConfig>.

=head1 DESCRIPTION

C<ClearCase::Config> contains all the environment and platform-specific
configuration settings and attributes for the version of ClearCase
running on the host system. Configuration settings may be obtained
by invoking the C<clearcase_config> function and passing it the name
of one or more attributes.

The set of ClearCase attribute names currently supported are as follows:

=over 4

=item C<ATRIA_HOME>

Path to Atria home directory

=item C<ATRIA_ADM>

Path to Atria administration directory

=item C<CLEARCASE_BIN>

Path to ClearCase F<bin> programs

=item C<CLEARCASE_ETC>

Path to ClearCase F<etc> programs

=item C<CLEARCASE_LOG>

Path to ClearCase log file directory

=item C<CLEARCASE_XN_SFX>

Extended naming symbol, usually "@@"

=item C<CLEARCASE_VIEWROOT>

View root directory, usually "F</view>"

=item C<CLEARTOOL>

The full pathname to the B<cleartool> executable program

=back

For the sake of convenience, the user may use the abbreviation C<CC_>
anywhere that C<CLEARCASE_> appears above; and C<CT_> may be used
anywhere that C<CLEARTOOL_> appears above. Any combination of uppercase
and lowercase characters may be used but they will always be mapped
to uppercase when computing the corresponding result.

=cut

#############################################################################

package ClearCase::Config;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

use Exporter;
use Carp;

@ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

@EXPORT = qw( );

@EXPORT_OK = qw( clearcase_config 
                 clearcase_progpath
                 clearcase_version
                 clearcase_patches
                 _clearcase_config
                 _clearcase_progpath
                 _clearcase_version
                 _clearcase_patches );
%EXPORT_TAGS = ( 'all' => [@EXPORT, @EXPORT_OK] );

$VERSION = 5.000_013;
require  5.004;    ## requires this Perl version or later

use Carp;
use File::Spec;
use ClearCase::OSUtil qw( MSWIN &devnull &progname &quote4cli );
require ClearCase::SiteConfig;

## Require the corresponding platform-specific config-hash
BEGIN {
    if (MSWIN) {
       require ClearCase::Config::MSWin32;
       import ClearCase::Config::MSWin32 qw(%CC_CONFIG);
    }
    else {
       require ClearCase::Config::Unix;
       import ClearCase::Config::Unix qw(%CC_CONFIG);
    }
}

## Initialize the other config-info hashes
use vars qw(%CC_ABBREVS %CC_PROGRAMS %CC_VERSIONS %CC_PATCHES);
%CC_ABBREVS  = ();
%CC_PROGRAMS = ();
%CC_VERSIONS = ();
%CC_PATCHES  = ();

my $DEVNULL = devnull;

##---------------------------------------------------------------------------
## Find all the executable programs
sub find_programs() {
   my $atriahome = $CC_CONFIG{'ATRIA_HOME'};
   ref($atriahome)  and  $atriahome = &{$atriahome};
   for ( qw(etc bin) ) {
      my $dir = File::Spec->catfile($atriahome, $_);
      ## Open the directory for reading
      opendir(DIRECTORY, $dir) or die "Can't open $dir for reading: $!";
      ## Read all the directory entries
      while ( defined($_ = readdir(DIRECTORY)) ) {
         next if (/^\.*$/);  ## ignore . and ..
         my $progname = progname($_);
         my $progpath = File::Spec->catfile($dir, $_);
         if (-x $progpath and ! -d _) {
            $CC_PROGRAMS{$progname} = $progpath;
         }
      }
      closedir(DIRECTORY);
   }
}

##---------------------------------------------------------------------------
## Find all the things we allow as abbreviations
sub find_abbrevs() {
   for (keys %CC_CONFIG) {
      next unless /^(ATRIA|CC|CLEAR(?:CASE|TOOL|GUIDE|QUEST)|MULTITOOL)(_\w+)?$/;
      my $abbrev = $_;
      if (/^CC/) {
         $abbrev =~ s/^CC/CLEARCASE/;
         $CC_ABBREVS{$abbrev} = $_;
      }
      if (/^ATRIA/) {
         $abbrev =~ s/^ATRIA/CLEARCASE/;
         $CC_ABBREVS{$abbrev} = $_;
      }
      if (/^CLEARCASE/) {
         $abbrev =~ s/^CLEARCASE/CC/;
         $CC_ABBREVS{$abbrev} = $_;
         $abbrev =~ s/^CC/ATRIA/;
         $CC_ABBREVS{$abbrev} = $_;
      }
      $abbrev =~ s{^(CLEAR|MULTI)(CASE|TOOL|GUIDE|QUEST)}
                  {substr($1,0,1) . substr($2,0,1)}e;
      $CC_ABBREVS{$abbrev} = $_;
   }
}

##---------------------------------------------------------------------------
## Parse -version output from a command. This is the function that
## computes a program version to populate into the %CC_VERSIONS
## and %CC_PATCHES hashes
sub parse_versions($) {
   my $cmdpath = shift;
   $cmdpath = qq("$cmdpath") if MSWIN;
   my $cmdline = "$cmdpath  -version < $DEVNULL 2> $DEVNULL";
   for (`$cmdline`) {
      chomp;
      s/^\s*\@\(\#\)\s*//;   ## remove any what(1) string prefixes
      s/\s+version\s+/ /;    ## remove the word "version" from the output
      next unless /^\s*(\S.*\S)\s+(\([^()]*\))\s*$/;
      my ($ver, $timestamp) = ($1, $2);
      my ($app, $rev) = split(" ", $ver, 2);
      if (defined $rev) {
         $CC_VERSIONS{$app} = $rev;
      }
      else {
         $CC_PATCHES{$app}  = $timestamp;
      }
   }
}

use vars qw(%hash);

##---------------------------------------------------------------------------
## Compute/expand a hash-value that is a code-ref to be the result
## of invoking the code (and dont forget to return the value)
sub lazy_config_val ( % $ ) {
   local *hash = shift;
   my $key = shift;
   my $val = $hash{$key};
   ref($val) and  $val = $hash{$key} = &{$val}();
   return $val;
}

##---------------------------------------------------------------------------
## Compute/expand all lazy values in a config-hash
sub expand_lazy_vals ( % ) {
   local *hash = shift;
   for ( grep { ref $hash{$_} } (keys %hash) ) {
       ref $hash{$_}  and  lazy_config_val(%hash, $_);
   }
}

#############################################################################

=head1 C<clearcase_config>

Queries ClearCase configuration attributes:

        my $scalar = clearcase_config($attr_name);
        my @array  = clearcase_config $attr1_name, $attr2_name;
	my %hash   = clearcase_config;

C<clearcase_config> takes zero or more configuration attribute names.
If no attribute names are given, then a hash is returned where the
attribute names are keys in the hash and the values are the corresponding
attribute settings. If one or more attribute names are given, then
the value of each corresponding attribute setting is returned (which
will result in an array if more than one name was given).

=cut

sub clearcase_config (@);

sub _clearcase_config (@) {
    if (@_ == 0) {
        ## Return the entire hash *after* expanding any lazy values
        expand_lazy_vals %CC_CONFIG;
        return %CC_CONFIG;
    }
    my @results = ();
    for (@_) {
        my $key = uc $_ || "";
        unless (exists $CC_CONFIG{$key}) {
            ## Check to see if its an accepted abbreviation
            %CC_ABBREVS  or  find_abbrevs();
            (exists $CC_ABBREVS{$key})  and  $key = $CC_ABBREVS{$key};
        }
        my $val = $CC_CONFIG{$key};
        ## If it's a reference, then it's a function-pointer to compute
        ## the actual result, which we then need to cache in our hash.
        ref($val)  and  $val = lazy_config_val(%CC_CONFIG, $key);
        push @results, $val;
    }
    (@_ == 1) ? $results[0] : @results;
}


#############################################################################

=head1 C<clearcase_progpath>

Obtains the program path(s) to the named ClearCase program(s).

        my $progpath  = clearcase_progpath($progname);
        my @pathnames = clearcase_progpath $prog1_name, $prog2_name;

C<clearcase_progpath> takes zero or more ClearCase program names.
If no program names are given, then a hash is returned where the
program names are keys in the hash and the values are the corresponding
program paths. If one or more program names are given, then
the pathname of each corresponding program is returned (which
will result in an array if more than one name was given).

=cut

sub clearcase_progpath (@);

sub _clearcase_progpath (@) {
    ## find executable programs if we havent already
    %CC_PROGRAMS  or  find_programs();
    (@_ == 0)  and  return %CC_PROGRAMS;
    my @results = ();
    for (@_) {
        push @results, $CC_PROGRAMS{$_};
    }
    (@_ == 1) ? $results[0] : @results;
}


#############################################################################

=head1 C<clearcase_version>

Version information for ClearCase applications:

        my $ct_ver = clearcase_version 'cleartool';
        my $cc_ver = clearcase_version 'ClearCase';
	my ($cmk_ver, $mvfs_ver) = clearcase_version qw(clearmake MVFS);

C<clearcase_version> takes the name of one or more ClearCase tools or
products and returns the version identifier corresponding to each one.
Valid tool/product names are:

=over 4

=item C<ClearCase>

installed version of ClearCase

=item C<MVFS>

installed version of the Multi-Version File System

=item C<cleartool>

installed version of B<cleartool> command

=item C<multitool>

installed version of B<multitool> command
(undefined if MultiSite is not installed)

=item C<xcleartool>

installed version of B<xcleartool> command

=item C<clearmake>

installed version of B<clearmake> command

=item C<view_server>

installed version of B<view_server>

=item C<vobrpc_server>

installed version of B<vobrpc_server>

=item C<db_server>

installed version of B<db_server>

=back

=cut


sub clearcase_version (@);

sub _clearcase_version (@) {
    return () unless @_;
    my @results = ();
    for (@_) {
        unless (exists $CC_VERSIONS{$_}) {
            ## Get program path to so we can invoke it with '-version'
            my $progpath = clearcase_progpath($_);
            unless (defined $progpath) {
               ## If no such program then assume its part of the output
               ## from 'cleartool -version'
               $progpath = $CC_CONFIG{'CLEARTOOL'}
                       unless (exists $CC_VERSIONS{'cleartool'});
            }
            parse_versions($progpath)  if (defined $progpath);
        }
        push @results, $CC_VERSIONS{$_};
    }
    (@_ == 1) ? $results[0] : @results;
}


#############################################################################

=head1 C<clearcase_patches>

Patch-level information for ClearCase installations:

        my $was_installed = clearcase_patches $patchname;
        my @installed_patches = clearcase_patches;

C<clearcase_patches> takes the name of one or more ClearCase patch-
names and returns a true value if each of the named patches has been
installed on the system. If no arguments are given, then it returns
a list of names of installed patches.

=cut

sub clearcase_patches (@);

sub _clearcase_patches (@) {
    ## Run 'cleartool -version' if we havent already
    parse_versions( $CC_CONFIG{'CLEARTOOL'} )
            unless (exists $CC_VERSIONS{'cleartool'});

    (@_ == 0)  and  return (keys %CC_PATCHES);    
    my @results = ();
    for (@_) {
        my $key = lc $_;
        $key =~ s/^(\d)/p$1/;
        $key =~ s/^(p\d)/clearcase_$1/;
        push @results, $CC_PATCHES{$key};
    }
    (@_ == 1) ? $results[0] : @results;
}


#############################################################################
## Now alias the exported names to the functions in SiteConfig if they
## exist, otherwise just alias them to the corresponding _xxxx function.
##
for ( map("clearcase_$_", qw(config progpath version patches)) ) {
    $ClearCase::Config::{$_} = $ClearCase::SiteConfig::{$_}
                                || $ClearCase::Config::{"_$_"};
}

#############################################################################

=head1 AUTHORS

Brad Appleton  F<E<lt>bradapp@enteract.comE<gt>>

=head1 SEE ALSO

ClearCase::SiteConfig, ClearCase::OSUtil

=cut

1;

__END__
