#############################################################################
# ClearCase/Config/MSWin32.pm -- Unix ClearCase Configuration Settings
#
# Copyright (C) 1995-2000 Brad Appleton. All rights reserved.
#
# This file is part of "ClearPerl" version 5 (ClearPerl5).
# This file/module is free software; you can redistribute it and/or
# modify it under the same terms as ClearPerl itself, as stated in
# the README file of the ClearPerl distribution.
#############################################################################

package ClearCase::Config::MSWin32;

use strict;
use Exporter;
use Carp;

use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw( %CC_CONFIG );

$VERSION = 5.000_013;
require  5.004;    ## requires this Perl version or later
use vars qw(%CC_CONFIG);
%CC_CONFIG = ();

#############################################################################


##-------------------------------------------------------------------------
## This sub is used to get information about ClearCase from the Windows
## Registry.  Supply a key path, and a value to look up at that path.
## It is assumed all paths are under HKEY_LOCAL_MACHINE.
##
## Returns undef on error.
##
## Note: We are using the Win32::Registry package that is part of the
## libwin32 package.  See http://www.perl.com/CPAN/authors/id/GSAR/
## for the latest version
##-------------------------------------------------------------------------
eval { require Win32::Registry };
import Win32::Registry unless $@;
sub query_winrgy( $ $ ) {
    return undef  unless (defined $main::HKEY_LOCAL_MACHINE);
    my ($path, $key) = @_;
    my ($registry, %values);
    unless ( $main::HKEY_LOCAL_MACHINE->Open($path, $registry) ) {
        warn "Error: Win32::Registry->Open: $!\n";
        return undef;
    }
    $registry->GetValues(\%values);
    $registry->Close();
    return $values{$key}->[2];
}

my $CC_WINRGY    = q(Software\Atria\ClearCase\CurrentVersion);
my $CC_MVFSPARMS = q(System\CurrentControlSet\Services\Mvfs\Parameters);

## Compute 'ATRIAHOME' at runtime as part of "lazy evaluation". Anything
## whose value depends on ATRIAHOME gets "set" when we compute ATRIAHOME.
## The first argument indicates the key of the value that needs to be
## returned.

sub lazy_ccpath($) {
    my $key = shift;
    my $atria_home = $ENV{ATRIAHOME}
                        || query_winrgy($CC_WINRGY,"ProductHome")
                        || q(c:\atria);
    my $atria_adm  = $atria_home . q(\adm);
    my $ccase_bin  = $atria_home . q(\bin);
    my $ccase_etc  = $atria_home . q(\etc);
    my $ccase_log  = $atria_home . q(\log);
    my $cleartool  = $ccase_bin  . q(\cleartool.exe);
    my $multitool  = $ccase_bin  . q(\multitool.exe);

    @CC_CONFIG{ qw(ATRIA_HOME ATRIA_ADM CLEARCASE_BIN CLEARCASE_ETC
                   CLEARCASE_LOG CLEARTOOL MULTITOOL) }
            = ( $atria_home, $atria_adm, $ccase_bin, $ccase_etc,
                $ccase_log, $cleartool, $multitool );

    return $CC_CONFIG{$key};
}


## NOTE: if a value in the hash is a code-ref, then it means we use
## lazy evaluation and dont execute the code-ref until the corresponding
## key has been used as an index, at which time we then invoke the
## code-ref and cache the result in the hash-value.

%CC_CONFIG = (
    ## Registry Path for ClearCase settings
    CLEARCASE_WINRGY    => $CC_WINRGY,
    CLEARCASE_MVFSPARMS => $CC_MVFSPARMS,
    CLEARCASE_XN_SFX    => sub {
                             query_winrgy($CC_MVFSPARMS, "xnsuffix") || q(@@)
                           },
    CLEARCASE_VIEWDRIVE => sub {
                             query_winrgy($CC_MVFSPARMS, "drive") || q(m:)
                           },
    CLEARCASE_VIEWROOT  => sub {
                             query_winrgy($CC_MVFSPARMS, "view_netname")
                                  || q(\\view)
                           },

    ## ClearCase paths
    ATRIA_HOME     => sub { lazy_ccpath 'ATRIA_HOME' },
    ATRIA_ADM      => sub { lazy_ccpath 'ATRIA_ADM' },
    CLEARCASE_BIN  => sub { lazy_ccpath 'CLEARCASE_BIN' },
    CLEARCASE_ETC  => sub { lazy_ccpath 'CLEARCASE_ETC' },
    CLEARCASE_LOG  => sub { lazy_ccpath 'CLEARCASE_LOG' },
    CLEARTOOL      => sub { lazy_ccpath 'CLEARTOOL' },
    MULTITOOL      => sub { lazy_ccpath 'MULTITOOL' },
);

1;

__END__
