#############################################################################
# ClearCase/ExecOpts.pm -- Execution options for cleartool
#
# Copyright (C) 1995-2000 Brad Appleton. All rights reserved.
#
# This file is part of "ClearPerl" version 5 (ClearPerl5).
# This file/module is free software; you can redistribute it and/or
# modify it under the same terms as ClearPerl itself, as stated in
# the README file of the ClearPerl distribution.
#############################################################################

package ClearCase::ExecOpts;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
use  5.004;    ## requires this Perl version or later
use  Exporter;
use  Carp;
use  File::Spec;

$VERSION = 5.000_013;
@ISA = qw( Exporter );

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

@EXPORT = qw();
@EXPORT_OK = qw( &execopts_parse &execopts_cmdline );
%EXPORT_TAGS = ( 'all' => [@EXPORT, @EXPORT_OK] );

use ClearCase::OSUtil qw( &redirect_cmd &quote4cli );

#############################################################################

=head1 NAME

ClearCase::ExecOpts - Execution options for cleartool

=head1 SYNOPSIS

  use ClearCase::ExecOpts;
  my $ct_opts = execopts_parse(
                       -status,
                       -chomping,
                       -execpath => '/usr/atria/local/bin'
                );
  print "return type is ", $ct_opts{'-returns'}, "\n";

  my $ct_cmdline = execopts_cmdline($ct_opts, 'lsco', '-short');
  system $ct_cmdline;

or

  my $ct_opts = new ClearCase::ExecOpts(
                       -status,
                       -chomping,
                       -execpath => '/usr/atria/local/bin'
                    );

  if (! $ct_opts->getopts('-output')) {
      $ct_opts->setopts('-output');
  }

  ## Now $ct_opts can be used with ClearCase::ClearTool
  ## or with a ClearTool object

=head1 REQUIRES/IMPORTS

perl5.004

=head1 EXPORTS

By default, I<nothing> is exported. The functions
B<execopts_parse> and B<execopts_cmdline>
may be exported upon request.

=head1 DESCRIPTION

ClearCase::ExecOpts lets users specify and customize the
execution options and the type of return value for a corresponding
B<ClearTool> object. Users may provide special keywords prefixed by
"-" or "+" to indicate the desired form of return-value for B<cleartool>
subcommand, or to indicate special run-time behavior for B<cleartool>
subcommands. These keywords are referred to as I<execution options>
for ClearCase::ClearTool objects and their derivations (abbreviated as
I<exec-options> or I<exec-opts>).

=head1 OPTIONS

The set of available execution options is given below. B<Please note> that
these options may use a prefix of "+" instead of the traditional "-" prefix
if desired.
This is so they may be easily distinguished from command-line options
to B<cleartool>(1) and its subcommands. Character-case is ignored when
recognizing an exec-option.

Some of the exec-options accept an argument; This is indicated by a "=>"
between the option-name and its argument.

=over 4

=item B<-status>

Return the exit status of the subcomand.

=item B<-success>

Return true if the exit-status of the command indicates success.
Return false otherwise.

=item B<-open>

Return an open file-handle (or tied handle) for reading the
output of the subcommand line-by-line.

=item B<-output>

Use the result of standard output as the return value
(this is the default). In a scalar context, I<all> of
the output is returned as a single string, with any
embedded newlines intact (with the possible exception
of the last newline). In an array context, a list of
output lines is returned. See the B<-chomping> and
B<-nochomping> exec-options regarding the disposition
of trailing newlines.

=item B<-chomping>

When standard output is to be returned, then this value controls
whether or not the returned result is to be "chomped" using the B<chomp>
built-in. If B<-chomping> is specified (which is the default), then in
a scalar context, the last newline (if any) of the resulting output
is removed. In an array context, the last newline of each line in the
returned list is removed.

=item B<-nochomping>

Returned unchomped results (trailing newlines intact) when output is
being captured.

=item B<-stdout> => ">I<filename>"|':stderr'|':stdnull'|C<undef>

Specify redirection for standard output of the subcommand.  If the string
':stdnull' is given, then it means output should be suppressed (neither
printed nor captured).  If the undefined value (C<undef>) is supplied, it
means the same thing as a value of ':stdnull'.  If the string ':stderr'
is given then output should be sent to the same place that diagnostic
error messages are currently being directed to. Otherwise, if the given
string begins with '>', '>>', or '|', then standard output is redirected
to the named file or command-filter.

=item B<-nostdout>

Equivalent to specifying C<-stdout =E<gt> undef>.

=item B<-stderr> => ">I<filename>"|':stdout'|':stdnull'|C<undef>

Specify redirection for standard diagnostic error output of the
subcommand. If the string ':stdnull' is given, then it means errors should
be suppressed (neither printed nor captured).  If the undefined value
(C<undef>) is supplied, it means the same thing as a value of ':stdnull'.
If the string 'stdout' is given then errors should be sent to the same
place that standard output currently being directed to. Otherwise,
if the given string begins with '>' or '>>' then error messages are
redirected to the named file.

=item B<-nostderr>

Equivalent to specifying C<-stderr =E<gt> undef>.

=item B<-abort> [ => I<msg-text>|I<code-ref>|I<hash-ref>|C<undef> ]

If status/success indicator is I<not> being returned, and if an non-zero
exit-status is returned by the subcommand, then automatically abort by
invoking croak/die with the specified message text, or else by invoking
the referred-to subroutine. If a hash-ref is given, then it may specify
both the message text I<and> the error subroutine by using C<-errmsg>
and C<-errsub> keyword (respectively). A value of C<undef>, which is
also the default, means do not abort! C<-abort> without an argument
requests abort on failure, with a standard message text.

NOTE THAT some commands sometimes use a status of '1' to indicate
a non-fatal alternate exit status. These are still non-zero and
B<-abort> will still abort when it encounters them. It is the caller's
responsibility to know whether or not they are invoking such a command
(e.g.: C<cleartool diff> is one such command) and to take any special
precautions (e.g.: using B<-noabort>) they require.

=item B<-noabort>

If status/success indicator is I<not> being returned, and if an unsuccessful
exit-status is returned by the subcommand, then I<do not> abort!
It is equivalent to specifying C<-abort =E<gt> undef>.

=item B<-intr> => I<value>

I<TO BE DETERMINED>

=item B<-nointr>

I<TO BE DETERMINED>

=item B<-execpath> => I<pathname>

The file-system path to the executable B<cleartool> command to use. If
the given I<pathname> corresponds to a directory, then it is assumed
that an executable file named F<cleartool> resides in that directory.
If no B<-execpath> is specified, then the default path to use will be
the value of the C<CLEARTOOL> attribute from B<ClearCase::Config>.

If the given pathname is a basename (i.e. a path without a directory
separator), then it will be resolved by the native operating system
(which usually means looking in the PATH environment variable). As a
I<special case>, if a pathname of "" (the empty string) is given, then
it is assumed to be "cleartool" (even if there is a subdirectory named
F<cleartool> om the current directory).

=item B<-name> => I<string>

The name of the object or instance or tool or application as it
should appear in any diagnostic messages.

=item B<-debug> [ => I<string>|C<undef>|"" ]

Set or unset a debugging flag for internal use.

=item B<-nodebug>

Equivalent to specifying C<-debug =E<gt> undef>.

=back

=cut

#############################################################################

=head1 EXPORTABLE FUNCTIONS

=head1 execopts_parse()

   ## return a hash-ref corresponding to the set of parsed options
   my $exec_opts = execopts_parse  @option_list;

Returns a reference to an associative array (hash) indicating all the
exec-option settings. The resulting hash will have the following keys
and their corresponding values:

=over 4

=item B<-returns>

Indicates the type of return value to use. The corresponding value for
this key will be a string set to one of "output", "status", "success",
or "open". These correspond respectively to the exec-options B<-output>,
B<-status>, B<-success>, and B<-open>.

=item B<-chomping>

The corresponding value will evalute to true if B<-chomping> is in
effect, and will evaluate to false otherwise (e.g. B<-nochomping>).

=item B<-stdout>

The corresponding value will be the empty string if no explicit value
was supplied for standard output redirection; Otherwise it will be the
value specified to the B<-stdout> exec-option, which will correspond to
a filename (preceded by '>' or '>>'), a command-filter (preceded by '|'),
standard-error-output (the string 'stderr'), or C<undef>
(meaning output is suppressed).

=item B<-stderr>

The corresponding value will be the empty string if no explicit value
was supplied for standard error redirection; Otherwise it will be the
value specified to the B<-stderr> exec-option, which will correspond to
a filename (preceded by '>' or '>>'), standard-output (the string 'stdout'),
or C<undef> (meaning error messages are suppressed).

=item B<-abort>

The corresponding value will be the empty string or C<undef> if no
automatic-aborting is to be performed; otherwise it will be a hash-ref
where the C<-errmsg> key indicates the error message to be printed
(if any), and the C<-errsub> key indicated the error-handler (if any)
to use instead of die/croak.

=item B<-intr>

The corresponding value will be B<undef> if no special interrupt 
processing was specified. If interrupts are to be disabled (B<-nointr>)
then the value will be the empty string.
If interrupts are to be fielded by an interrupt-handler, then the value
will be I<[TO BE DETERMINED]> the subroutine that handles the interrupt.

=item B<-execpath>

The corresponding value will be the full pathname to the executable
that is used to invoke B<cleartool> subcommands.

=item B<-name>

The corresponding value will be the name of the object or instance or
application or tool as it should appear in any diagnostic messages.

=item B<-debug>

The corresponding value will be the flag or level for internal debugging
(or undef or "" if debugging is disabled).

=back

=cut

use vars qw(%execopts);

sub execopts_parse (@) {
   my @opts = @_;
   local $_ = '';
   my $execopts_ref = {};
   local *execopts  = $execopts_ref;

   while (@opts > 0) {
      $_ = shift(@opts);
      my $reftype = ref;
      if ($reftype =~ /^(?:HASH|ClearCase::ExecOpts)$/) {
         ## Just append the given hash and move on
         %execopts = (%execopts, %$_);
         next;
      }
      if ($reftype eq 'ARRAY') {
         ## Push the array on the front of @opts and start processing it
         unshift @opts, @{$_};
         next;
      }
      s/^\W//;     ## remove leading '-' or '+' (or non-punct)
      $_ = lc $_;  ## convert to lowercase
      if ( /^((returns)|output|status|success|open)/ ) {
         $execopts{'-returns'} = $2 ? shift(@opts) : $1;
      }
      elsif ( /^(no)?(chomping)$/ ) {
         $execopts{"-$2"} = (length($1) == 0);
         ## See if arg of was given as undef|""|0|1|:(std)?(out|err|all)
         $execopts{"-chomping"} = shift @opts
             if (@opts and (!defined $opts[0]) || (!length  $opts[0]) ||
                 $opts[0] =~ /^(?:[01]|:(?:(?:std)?(?:out|err|all)))$/ );
      }
      elsif ( /^(no)?(stdout|stderr|name|debug|abort|intr)$/ ) {
         if ( $_ eq 'abort'  ||  $_ eq 'debug'  and
              ( @opts == 0 ||
                defined( $opts[0] ) && $opts[0] =~ /^[-+]/ ) ) {
            ## specified -abort or -debug WITHOUT an argument
            $execopts{"-$_"} = 1;
         }
         else {
            $execopts{"-$2"} = $1 ? undef : shift(@opts);
         }
      }
      elsif ( $_ eq 'execpath' ) {
         my ($key, $pname, $tool) = ("-$_", shift @opts, 'cleartool');
         if ( defined($pname) and length($pname) ) {
            $execopts{$key} = (-d $pname) ?
                    File::Spec->catfile($pname, $tool) : $pname;
         }
         else {
            $execopts{$key} = $tool;
         }
      }
      else {
        die( "no such exec-option ($_)\n" );
        return undef();
      }
   } ## while

   return $execopts_ref;
}


#############################################################################

=head1 execopts_cmdline()

   ## return a string for invoking the corresponding cleartool subcommand
   my $cmd_str = execopts_cmdline($exec_opts, 'lsco -short');

   ## Same as above but return an array instead
   my $cmd_argv = execopts_cmdline($exec_opts, 'lsco', '-short');

Returns a string (or an array when in an array context) corresponding
to the command-line invocation to use for the given set of exec-options
and the given (sub)command and arguments.

=cut

sub execopts_cmdline (@) {
   ## NOTE: This routine is extremely non-trivial. The main thing it
   ## has to deal with is i/o redirection via the -stdout and -stderr
   ## exec-options. It has to insert any '>outfile' or '2>errfile'
   ## to the given command-line. The tricky part is that if '-stdout'
   ## is to go to 'stderr' or vice-versa, it needs to assign the '1>...'
   ## and/or '2>...' in the correct order (the same as the order given).
   ## Of course, it also needs to get the proper -execpath to use for
   ## cleartool(1).  One day it may need to deal with pipes+filters.
   ##

   ## Get the exec-options
   local *execopts = shift;
   my $cleartool = $execopts{'-execpath'};
   my @argv = ($cleartool, quote4cli @_);

   my @redirects = redirect_cmd(\%execopts);

   ## If stdout is a pipeline, or if stdout -> stderr, then
   ## stderr should come *before* stdout; Otherwise put
   ## stdout *before* stderr
   ## 
   my $reverse_it = ( (@redirects > 1)
                      and ($execopts{'-stdout'} =~ /^\s*\|/ 
                           or  lc($execopts{'-stderr'}) eq ':stdout') );
   push @argv, ($reverse_it) ? reverse @redirects : @redirects;

   ## Always return a scalar - system is clever enough
   return join(" ", @argv);
}

#############################################################################

=head1 CONSTRUCTION/CREATION METHODS

=head1 new()

Constructs a new B<ClearCase::ExecOpts> object:

    ## Create a new ExecOpts object for cleartool subcommands
    my $ct_opts1 = ClearCase::ExecOpts->new (
                       -status,
                       -chomping,
                       -execpath => '/usr/atria/local/bin'
                   );

    ## Same as above, but used a different syntax
    my $ct_opts2 = new ClearCase::ExecOpts (
                          -status,
                          -chomping,
                          -execpath => '/usr/atria/local/bin'
                       );

=cut

sub new {
    ## Determine if we were called via an object-ref or a classname
    my $this = shift;
    my $class = ref($this) || $this;
    ## Any remaining arguments are treated as initial values for the
    ## hash that is used to represent this object.
    my $self = execopts_parse(@_);
    ## Bless ourselves into the desired class and perform any initialization
    bless $self, $class;
    $self->initialize();
    return $self;
}

sub initialize {
    my $self = shift;
    my %opts = %{ $self };
}

#############################################################################

=head1 OBJECT INSTANCE METHODS

=head1 getopts()

Queries exec-option settings:

    if ($ct_opts->getopts('-chomping')) {
       print "removing trailing newlines output\n";
    }
    my $returns = $ct_opts->getopts('-returns');
    print "-returns is $returns\n";
    my ($abort, $execpath) = $ct_opts->getopts('-abort', '-execpath');
    my %optvals = $ct_opts->getopts();

B<getopts()> will return the value of the specified exec-option. If more
than one exec-option is specified, then a list of corresponding exec-option
values is returned. If no exec-options are specified, then a hash, of the
same form returned by B<execopts_parse>, is returned to the caller.

The arguments to B<getopts> should be the names of one or more exec-options.
Since B<getopts> queries exec-option values, none of the exec-option names
take any arguments. In particular, this means that:

=over 4

=item *

B<-returns> corresponds to a value that matches
C</^(output|status|success|open)$/>.

=item *

Any other exec-option that takes an argument returns the corresponding
argument value.

=item *

Any other exec-option that does I<not> take an argument returns a boolean
result, based on whether or not that option is set.

=back

So if both B<-returns> and B<-chomping> are specified to B<getopts>, this
would correspond to I<two> values: the first would be a string and the
second would be a boolean result.

=cut

sub getopts {
   my $self = shift;
   local *execopts = $self;
   return  %execopts         if (@_ == 0);
   return  $execopts{$_[0]}  if (@_ == 1);
   return  @execopts{@_}     if (@_ > 1);
}

##---------------------------------------------------------------------------

=head1 setopts()

Modifies exec-option settings:

    $ct_opts->setopts('-intr' => undef);
    if ($ct_opts->getopts('-chomping')) {
       $ct_opts->setopts('-nochomping');
    }
    $ct_opts->setopts('-returns' => 'output');
    $ct_opts->setopts('-output', '-execpath' => '/my/local/path/ct');
    my %optvals = $ct_opts->getopts();
    $ct_opts->setopts( \%optvals );

B<setopts()> will set the specified exec-options. It accepts the same
argument syntax as the B<new> constructor. In addition, if the first
parameter given is a reference to a list or a hash, then B<setopts>
assumes the corresponding hash is of the same form that is returned by
B<execopts_parse>, and uses it to reset all the exec-options. This
may be useful when you want to restore the exec-options to their original
state after modifying them:

    my $save_opts = $ct_opts->getopts;
    $ct_opts->setopts( qw{ -status -nocapture -nointr } );
    ## ... do some things using the current settings

    ## now restore the old settings
    $ct_opts->setopts( \%save_opts );

It should be noted that B<setopts> does not return any particular value.

=cut

sub setopts {
   my $self = shift;
   local *execopts = $self;
   my $opts = execopts_parse(@_);
   %execopts = (%execopts, %$opts);
}

#############################################################################

=head1 AUTHORS

Brad Appleton  F<E<lt>bradapp@enteract.comE<gt>>

=head1 SEE ALSO

ClearCase::ClearTool,
ClearCase::Config

=cut

1;

__END__
