#############################################################################
# ClearCase/OSUtil.pm -- OS platform-specific file/io functions
#
# Copyright (C) 1999-2000 David Boyce. All rights reserved.
#
# Modified (with permission) for ClearPerl by Brad Appleton
#
# This file is part of "ClearPerl" version 5 (ClearPerl5).
# This file/module is free software; you can redistribute it and/or
# modify it under the same terms as ClearPerl itself, as stated in
# the README file of the ClearPerl distribution.
#############################################################################

=head1 NAME

ClearCase::OSUtil - OS-independent file and i/o utilities

=head1 SYNOPSIS

    use ClearCase::OSUtil;

or

    use ClearCase::OSUtil @import_list;

=head1 DESCRIPTION

Provides a set of filename and i/o utility functions for ClearCase-Perl
programming. Several constants are provided for dealing with OS-dependent
paths and delimiters, along with and several functions for manipulating
file pathnames in an OS-independent manner. Several constants and
functions for basic input/output handling and redirection are also
provided.

=head1 EXPORTS

By default, I<nothing> is exported. The following subroutine names are
exported upon request:

  splitpathlist, joinpathlist, progname, progpath,
  devnull, tmpdir, tmpname, redirect_cmd, redirect_fh,
  env4cli, quote4cli, sendmsg

In addition, the following constants and read-only scalars may be
exported upon explicit request:

  MSWIN, $PATHLISTSEP

=cut

#############################################################################

package ClearCase::OSUtil;

use strict;
use Exporter;
use Carp;
use File::Basename;
use File::Spec;

use vars qw( $VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS
             $PATHSEP $PATHLISTSEP $STDNULL $TMPDIR
           );
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw( &progname &progpath
                 &devnull &tmpdir &tmpname
                 &splitpathlist &joinpathlist
                 &redirect_cmd &redirect_fh
                 &env4cli &quote4cli &sendmsg
                 MSWIN $PATHLISTSEP
               );
%EXPORT_TAGS = ( 'all' => [@EXPORT, @EXPORT_OK] );

$VERSION = 5.000_013;
require  5.004;    ## requires this Perl version or later

#############################################################################

=head2 CONSTANTS

=over 6

=item C<MSWIN>

The constant C<MSWIN> is set to 1 (but not automatically
exported) iff the platform is Windows.

=item C<$PATHLISTSEP>

The read-only scalar C<$PATHLISTSEP> is set (but not automatically
exported) to the pathname-list separator between pathnames within a
search-path string.

=back

=cut

#############################################################################

use constant MSWIN  => (lc $^O  eq  'mswin32');

if (MSWIN) {
    *PATHSEP     = \( "\\" );
    *PATHLISTSEP = \( ";" );
    *STDNULL     = \( 'NUL' );
}
else {
    *PATHSEP     = \( "/" );
    *PATHLISTSEP = \( ":" );
    *STDNULL     = \( '/dev/null' );
}

#############################################################################

=head2 FILE/PATH MANIPULATION FUNCTIONS

The following functions may be used to manipulate pathnames and filenames.

=over 4

=cut

##-------------------------------------------------------------------------

=item B<joinpathlist>

Join an array of pathnames into a delimited list pathnames. Example:

   $ENV{PATH} = joinpathlist( /usr/atria/bin /usr/atria/etc );

=cut

sub joinpathlist(@) {
    return  join($PATHLISTSEP, @_);
}

##-------------------------------------------------------------------------

=item B<splitpathlist>

Split a delimited list of pathnames into an array of pathnames. Example:

    my @cmdpaths = splitpathlist $ENV{PATH};

=cut

sub splitpathlist(@) {
    return  map { split($PATHLISTSEP) } @_;
}

##-------------------------------------------------------------------------

=item B<progname>

Return the invocation name of a program (stripping off any
OS-dependent file-extension for executable programs). Example:

   my $ctpath = joinpath qw(usr atria bin cleartool.exe);
   my $ctname = progname $ctpath;  ## strips the '.exe' extension

On Unix platforms , no suffix-stripping is performed and the function
simply returns its arguments. On MSWin32 platforms, the extensions in
the C<PATHEXT> environment variable are stripped.

=cut

sub progname(@) {
    ## If this isnt Win32, dont need to do anything
    MSWIN  or  return (@_ == 1) ? $_[0] : @_;

    ## Otherwise this is Win32, strip off any command/execution suffixes
    my $pathexts = $ENV{PATHEXT} || "";

    ## Need to make a regular expression to match all the suffixes
    my @suffixes = map { quotemeta } split($PATHLISTSEP, $pathexts);
    return  @_   unless @suffixes;
    my $regex = '(?i:' . join('|', @suffixes) . ')';

    my @results = ();
    for (@_) {
       my ($basename,$dirpath,$suffix) = fileparse($_,$regex);
       $dirpath = ""  unless (index($_, $dirpath) == 0);
       push @results, $dirpath.$basename;
    }
    return (@_ == 1) ? $results[0] : @results;
}

##-------------------------------------------------------------------------

=item B<progpath>

Find the path to the given program. Searches the PATH environment variable
if the given name is not an absolute path. If a corresponding executable
pathname is found, then it is returned; otherwise C<undef> is returned.

   my $progname = 'cleartool';
   my $ctpath = progpath $ctprog;

On MSWin32 platforms, if the given program does not have a suffix, then
the C<PATHEXT> environment variable is used to find the suffix that will
be used.

=cut

sub progpath(@) {
    @_ == 0  and  return undef;
    my @results = ();

    ## Split-up environment variables now
    my @pathexts = MSWIN ? splitpathlist(lc $ENV{PATHEXT}) : ();
    my @pathdirs = splitpathlist($ENV{PATH});

    my ($prog, $dir, $ext, $progpath);
    for $prog (@_) {
        next unless (defined $prog  and  length $prog);
        ## Compute possible extensions if $prog doesnt already have one
        my @exts = (MSWIN and !/\.[^.]{1,3}$/) ? @pathexts : ("");
        ## Determine if $prog is an absolute path
        my $is_abspath = MSWIN ? $prog =~ m{^(?:[A-Za-z]:)?[/\\]}
                               : $prog =~ m{^/};
        ## Search through pathdirs unless this is an absolute path
        for $dir (@pathdirs) {
            my $pname = $is_abspath ? $prog : File::Spec->catfile($dir, $prog);
            for $ext (@exts) {
                $progpath = $pname.$ext;
                last if -x $progpath;  ## found
                $progpath = "";  ## not found
            }
            ## If found, or absolute path, we're done
            last  if (length($progpath) or $is_abspath);
        }
        ## append the path if we found it and its executable, otherwise undef
        push @results, length($progpath) ? $progpath : undef;
    }
    return (@_ == 1) ? $results[0] : @results;
}

##########################################################################

=back

=head2 INPUT/OUTPUT HANDLING AND REDIRECTION

=over 4

=cut 

###########################################################################

##-------------------------------------------------------------------------

=item B<devnull>

Returns the name of the "null input/output device" used on the
command-line for redirecting input or output to the bit-bucket.

   my $stdnull = devnull;
   ## should be '/dev/null' on Unix and 'NUL' on MSWin32

I<Note> that if you are using Perl version 5.005_58 or later, you should
use B<File::Spec::Functions::devnull> instead!

=cut

sub devnull() {
    return $STDNULL;
}

##-------------------------------------------------------------------------

=item B<tmpdir>

Returns the pathname of the directory in which to create temporary files.
Example:

    my $tmpd = tmpdir;

I<Note> that if you are using Perl version 5.005_58 or later, you should
use B<File::Spec::Functions::tmpdir> instead.

=cut

sub tmpdir() {
    if (MSWIN) {
        my $SYSDRIVE = $ENV{SYSTEMDRIVE} || "";
        return  $ENV{TEMP} || $ENV{TMP} ||
                (-d "$SYSDRIVE/temp" ? "$SYSDRIVE/temp" : $SYSDRIVE);
    }
    else {
        return $ENV{TMPDIR} || $ENV{TEMPDIR} ||
               (grep {-d} qw(/tmp /usr/tmp))[0] || q(/var/tmp);
    }
}
*TMPDIR = \(tmpdir);

##-------------------------------------------------------------------------

=item B<tmpname>

The tmpname() function generates a name for a temporary file which
doesn't yet exist.  It makes no pretense of being atomic, so it's
conceivable (though exceedingly unlikely) that the generated filename
could be taken between the time it's provided and the time it's
opened. Example:

    my $tmpf = tempfile('foo');

The optional parameter ('foo') becomes a filename extension. It can
be used to make the tempfile's purpose more mnemonic; uniq-ification
comes from combining the pid with a random number and does not depend
on the parameter.

=cut

sub tmpname($) {
   my $custom = shift;
   (my $tmpd = $TMPDIR) =~ s%\\%/%g;
   (my $pkg = lc __PACKAGE__) =~ s/:+/-/g;
   while (1) {
      my $tmpf = join('.', "$tmpd/$pkg", $$, int(rand 10000));
      $tmpf .= $custom ? ".$custom.tmp" : '.tmp';
      return $tmpf  unless (-f $tmpf);
   }
}

##-------------------------------------------------------------------------

=item B<redirect_cmd>

Takes hash-ref as an argument that specifies redirection
keys like "-stdout" and "-stderr" (with values as specified in
ClearCase::ExecOpts.pm) and returns a list of command-line arguments to
perform i/o redirection for an arbitrary command. (Also understands a
key of "-stdin" with a value of C<undef> or ":stdnull".) Example:

   my $cmd = "sed -e 's/foo/bar/g";
   my @ioargs = redirect_cmd( {-stdout => ':stderr',
                               -stderr => ':stdnull'} );
   system "$cmd @ioargs";

=cut

   ## Need something to map redirection exec-opts to command-line
   ## directives for the shell. An 'undef' corresponds to a key of
   ## 'stdnull'. The following should work for Unix, but not NT.
my %REDIRECT = (
    'stdin'  => '<',
    'stdout' => '>',
    'stderr' => '2>',

    'stdin2stdin'    => '',
    'stdin2stdnull'  => '<'.$STDNULL,

    'stdout2stdout'  => '',
    'stdout2stderr'  => '>&2',
    'stdout2stdnull' => '>'.$STDNULL,

    'stderr2stderr'  => '',
    'stderr2stdout'  => '2>&1',
    'stderr2stdnull' => '2>'.$STDNULL,
);

use vars qw(%hash);

sub redirect_cmd( % ) {
   local *hash = shift;
   my @redirects = ();
   for ( qw(stdin stdout stderr) ) {
      next  unless ( exists $hash{"-$_"} );
      my $dest = $hash{"-$_"} || ':stdnull';
      $dest =~ s/^\s*//;  ## trim leading whitespace
      my $redirected = '';
      if ( lc($dest) =~ /^:(stdout|stderr|stdnull)$/ ) {
         $redirected = $REDIRECT{$_.'2'.$1};
      }
      elsif ( substr($dest,0,1) eq '|' ) {
         ## this is a pipe, dont use any leading '>'
         $redirected = $dest;
      }
      else {
         ## to a file - remove leading '<' or '>' if its there
         $dest =~ s/^[<>]//;
         $redirected = $REDIRECT{$_} . $dest;
      }
      push @redirects, $redirected;
   }
   return @redirects;
}


##-------------------------------------------------------------------------

=item B<redirect_fh>

Takes a hash-ref as its argument, with keys specifying "-stdout" or
"-stderr" (with values of "ON" or "OFF") and enables or disables
stdout/stderr filehandles as specified.

Sometimes you need to suppress output to stdout or stderr, or both. In
many cases the need to do so is only temporary. All of this is do-able
in basic Perl but it can be awkward; for instance here's a sequence
which saves, closes, and restores STDOUT:

    use vars '*SAVEOUT';
    open(SAVEOUT, ">&STDOUT");
    close(STDOUT);
    system qw(date); # this output is suppressed
    open(STDOUT, ">&SAVEOUT");

Messy, eh? People often try to do it this way for child processes:

    system("date >/dev/null 2>&1");

but this requires the use of an intermediate shell, which is to be avoided
where possible even on Unix platforms for various reasons, and can seriously
complicate Unix/Windows interoperability.

The redirect_fh() function encapsulates all of these complexities without
requiring a shell for child processes. The equivalent of the save/restore
sequence above is:

    redirect(-stdout => 'OFF');
    system qw(date); # this output is suppressed
    redirect(-stdout => 'ON');

You can handle STDERR the same way, or even both together e.g.:

    redirect(-stdout => 'OFF', -stderr => 'OFF');

=cut 

{
my %FILEMAP = map { ($_ => "") } qw(STDOUT STDERR);
sub redirect_fh(@) {
   # Stash these away at first use for potential future use, e.g. debugging.
   open(SAVE_STDOUT, '>&STDOUT') unless defined fileno(SAVE_STDOUT);
   open(SAVE_STDERR, '>&STDERR') unless defined fileno(SAVE_STDERR);

   ## We are more forgiving than the PODs specify. We claim to require
   ## a hash, but we allow for a list. And unlike redirect_cmd, we
   ## allow '-stdout' and '-stderr' to be any combination of upper/lower
   ## case with or without a leading '-' prefix. This is for backward
   ## compatibility with a previous incarnation of this function.

   my @actions = map { ref($_) ? %{$_} : $_ } @_;  ## expand any hash-refs

   while (@actions) {
      (my $stream = uc shift @actions) =~ s/^[-+]//;
      my $state = shift @actions || 'OFF';
 
      if ($state =~ /^[<>]/) {
         ## Remove leading redirect-char
         $state = substr($state, 1);
      }
      else {
         $state = uc $state;
         if ($state ne 'OFF'  and  $state ne 'ON') {
	    print SAVE_STDERR "unrecognized state $state for $stream\n";
	    next;
         }
      }
      if ($stream ne 'STDOUT'  and  $stream ne 'STDERR') {
	 print SAVE_STDERR "unrecognized stream $stream\n";
	 next;
      }

      no strict "refs";  ## allow STDOUT as a symbolic-ref
      if (length $FILEMAP{$stream}) {
         close($stream);
         $FILEMAP{$stream} = "";
      }
      my ($hide_fh, $save_fh) = ("HIDE_$stream", "SAVE_$stream");
      if ($state eq 'ON') {
	 open($stream, ">&$hide_fh");
	 select $stream;
         #print "Opened $stream ...\n";
      }
      elsif ($state eq 'OFF'  and  defined fileno($stream)) {
	 open($hide_fh, ">&$stream") unless defined fileno($hide_fh);
	 close($stream);
	 select $save_fh;
         #print "Closed $stream ...\n";
      }
      else {
         ## redirect to named file
	 open($stream, ">$state");
	 select $stream;
         $FILEMAP{$stream} = $state;
         #print "Redirected $stream to $state ...\n";
      }
   }
}
}

##-------------------------------------------------------------------------

=item B<env4cli>

Takes the the name of one or more environment variables and returns
the name in a format that can be interpolated by the command line.
Example:

   my $home_ev = env4cli qw(HOME);  ## returns '$HOME' on Unix,
                                    ## returns '%HOME%' on Win32
   system "echo $home_ev";

=cut

sub env4cli ($ ; @) {
   my @results = ();
   for (@_) {
      push @results, (MSWIN ? '%'.$_.'%' : '$'.$_);
   }
   return (@_ == 1) ? $results[0] : @results;
}

##-------------------------------------------------------------------------

=item B<quote4cli>

Quote words if necessary so that the command-line interpreter treats
them as a single word (token). If one of the given strings contains
whitespace, quotes will be needed. And on MSWin32 using CMD.EXE and
COMMAND.COM, arguments to functions like system() may need an extra level
of quoting if they are already quoted.

=cut

sub quote4cli (@) {
   my @argv = @_;
   for (@argv) {
      ## Skip arg if already quoted ...
      next if m%^".*"$%s;
      ## Special case - turn internal newlines back to literal \n on Win32
      s%\n%\\n%gs if MSWIN; 
      ## Skip if contains no special chars.
      next unless m%[^-=:_.\w\\/]% || tr%\n%%;
      ## Special case - leave things that look like redirections alone.
      next if /^\d?(?:<{1,2})|(?:>{1,2})/;
      ## Now quote embedded quotes ...
      $_ =~ s%(\\*)"%$1$1\\"%g;
      ## quote a trailing \ so it won't quote the quote (!) ...
      s%\\{1}$%\\\\%;
      ## and last the entire string.
      $_ = qq("$_"); 
   }   
   return @argv;
}

##-------------------------------------------------------------------------

=item B<sendmsg>

The subroutine sendmsg() can be used to deliver an email message
using the Net::SMTP module. One of its features is that it includes
Net::SMTP only if it's installed; thus you can use the other interfaces
of ClearCase::OSUtil even if Net::SMTP is not installed.  Note: this
function assumes that your SMTP module has been initialized correctly
with a default SMTP host (see <perl-install-dir>/.../Net/Config.pm for
details). It works from both Unix and NT machines without requiring
any configuration on client systems and is thus ideal for a networked
interop environment. Example:

    sendmsg([qw(addr1 addr2 addr3)], 'Subject here', @MsgBody);

It takes a ref to a list of email addresses followed by a string scalar
containing the subject. Any remaining parameters are used as the body
of the message. Returns true on successful delivery of msg (to the
SMTP host).

=cut

##-------------------------------------------------------------------------

sub sendmsg( $ $ @ ) {
   my($r_to, $subj, @body) = @_;
   # If no mailto list, no mail.
   return 0 unless ($r_to);
   return 0 if (ref($r_to) && ! @$r_to);
   # Only drag Net::SMTP in at runtime since it's not core perl.
   eval { require Net::SMTP; import Net::SMTP qw() };
   return 0 if $@;

   my $smtp = Net::SMTP->new;	## assumes Net::SMTP Config.pm is set up right
   return 0 unless defined $smtp;
   local $^W = 0; ## hide a spurious warning - from deep in Net::SMTP maybe??
   $smtp->mail($ENV{CLEARCASE_USER} || $ENV{LOGNAME} || $ENV{USERNAME}) &&
      $smtp->to(ref($r_to) ? @$r_to : $r_to) &&
      $smtp->data() &&
      $smtp->datasend("To: ".join(' ', ref($r_to) ? @$r_to : $r_to)."\n") &&
      $smtp->datasend("Subject: $subj\n") &&
      $smtp->datasend("\n") &&
      $smtp->datasend(@body) &&
      $smtp->dataend() &&
      $smtp->quit;
}

1;

#############################################################################

=back

=head1 PORTING

This package is known to work fine on Solaris 2.5.1-7 and Windows NT
4.0SP3-5.  As these two platforms are quite different, this should take
care of any I<significant> portability issues, but please send reports
of tweaks needed for other platforms to the address(es) below.

=head1 AUTHORS

David Boyce E<lt>dsb@world.std.comE<gt> (original version)
Brad Appleton E<lt>bradapp@enteract.comE<gt> (modified for Clearperl)

Copyright (c) 1999-2000 David Boyce. All rights reserved.  This perl program
is free software; you may redistribute it and/or modify it under the
same terms as Perl itself.

=head1 SEE ALSO

Net::SMTP, ClearCase::ExecOpts.

=cut

1;

__END__
